@@ -79,8 +79,8 @@ sub builder(&) {
 
     my $app = $block->();
     if ($mount_is_called && $app ne $urlmap) {
-        Carp::carp("You used mount() in builder block but the last line (app) isn't using mount().\n" .
-                       "This causes all mount() mappings ignored. See perldoc Plack::Builder for details.");
+        Carp::carp("You used mount() in a builder block, but the last line (app) isn't using mount().\n" .
+                       "This causes all mount() mappings to be ignored. See perldoc Plack::Builder for details.");
     }
 
     $self->to_app($app);
@@ -194,21 +194,29 @@ Plack::Builder has a native support for L<Plack::App::URLMap> with C<mount> meth
 See L<Plack::App::URLMap>'s C<map> method to see what they mean. With
 builder you can't use C<map> as a DSL, for the obvious reason :)
 
-B<Note>: Once you use C<mount> in your builder code, you have to use
+B<NOTE>: Once you use C<mount> in your builder code, you have to use
 C<mount> for all the paths, including the root path (C</>). You can't
 have the default app in the last line of C<builder> like:
 
+  my $app = sub {
+      my $env = shift;
+      ...
+  };
+
   builder {
       mount "/foo" => sub { ... };
-      sub {
-          my $env = shift; # THIS DOESN'T WORK
-      };
+      $app; # THIS DOESN'T WORK
   };
 
 You'll get warnings saying that your mount configuration will be
-ignored.  Instead you should use C<< mount "/" => ... >> in the last
+ignored. Instead you should use C<< mount "/" => ... >> in the last
 line to set the default fallback app.
 
+  builder {
+      mount "/foo" => sub { ... };
+      mount "/" => $app;
+  }
+
 =head1 CONDITIONAL MIDDLEWARE SUPPORT
 
 You can use C<enable_if> to conditionally enable middleware based on
